/*
 * Simulated data logger sketch
 */

#include "RingBuffer.h"

static const int  LED      = 13;       // for diagnostic
             bool saveFlag = false;

// Serial communication configuration:
#define DEBUG_PORT_TYPE HardwareSerial
#define DEBUG_PORT Serial

#define SAVE_TO_SD
#ifdef SAVE_TO_SD
  #include <SdFatYield.h>

  // SD card and saving configuration:
  SdFat SD;
  File logfile;
  const int chipSelect = 53;               // CS pin for the SD card board

  uint32_t lastFlushTime;

  const uint32_t FLUSH_PERIOD    = 60000UL; // ms
        uint8_t  flushCount      = 0;

  const uint32_t LRECORD_PERIOD  = 10000UL; // us
        uint32_t lastLrecord     = 0UL;
        uint32_t loggingTime     = 0UL;
#endif

bool     readDuringYield = false;
uint32_t yieldCount = 0UL;

uint32_t currentTime;

#define BUFFERING_TIME (750000UL) // us, longest SD write time

//------------------------------------------------------

uint32_t lastWSSread, lastSteeringRead, lastIMUread;

bool     IMUreading      = false;

struct IMUreading_t
  {
    uint32_t t;         // START time
    uint32_t dt2;       // imu.available() time - START time
  };

const uint32_t STEERING_PERIOD  = 10000UL; // us
const uint32_t WSS_PERIOD       = 10000UL; // us
const uint32_t IMU_PERIOD       = 10000UL; // us

const uint16_t WSS_RB_SIZE      = BUFFERING_TIME / WSS_PERIOD;
const uint16_t IMU_RB_SIZE      = BUFFERING_TIME / IMU_PERIOD;
const uint16_t STEERING_RB_SIZE = BUFFERING_TIME / STEERING_PERIOD;

RingBuffer< uint32_t     , WSS_RB_SIZE     > WSSreadings;
RingBuffer< IMUreading_t , IMU_RB_SIZE     > IMUreadings;
RingBuffer< uint32_t     , STEERING_RB_SIZE> SteeringReadings;

#define GPS_PERIOD (50000UL) // us, set by a configuration command
#define GPS_RB_SIZE (BUFFERING_TIME/GPS_PERIOD)
RingBuffer< uint32_t     , GPS_RB_SIZE     >  GPSreadings; // just a timestamp for simulation
uint32_t lastGPStime;

//------------------------------------------------------

void readGPS()
{
  currentTime = micros();

  if (currentTime - lastGPStime >= GPS_PERIOD) {
    GPSreadings.write( currentTime );

    uint8_t steps = 0;
    do {
      lastGPStime += GPS_PERIOD;
      steps++;
    } while (currentTime - lastGPStime >= GPS_PERIOD);

    if (steps > 1) {// then sample missed?
      DEBUG_PORT.print( 'G' );
      DEBUG_PORT.print( currentTime );
      DEBUG_PORT.print( '=' );
      DEBUG_PORT.println( steps );
    }
  }
} // readGPS

//------------------------------------------------------

void readWSS()
{
  currentTime = micros();

  if (currentTime - lastWSSread >= WSS_PERIOD) {

    WSSreadings.write( currentTime );

    uint8_t  steps    = 0;
    do {
      lastWSSread += WSS_PERIOD;
      steps++;
    } while (currentTime - lastWSSread >= WSS_PERIOD);

    if (steps > 1) {// then sample missed?
      DEBUG_PORT.print( 'W' );
      DEBUG_PORT.print( currentTime );
      DEBUG_PORT.print( '=' );
      DEBUG_PORT.println( steps );
    }
  }
} // readWSS

//------------------------------------------------------

void readIMU()
{
  currentTime = micros();

  if (!IMUreading) {

    // Is it time to take a new reading?
    if (currentTime - lastIMUread >= IMU_PERIOD) {

      //  Start the I2C transfer...

      uint8_t steps = 0;
      do {
        lastIMUread += IMU_PERIOD;
        steps++;
      } while (currentTime - lastIMUread >= IMU_PERIOD);

      if (steps > 1) {// then sample missed?
        DEBUG_PORT.print( 'I' );
        DEBUG_PORT.print( currentTime );
        DEBUG_PORT.print( '=' );
        DEBUG_PORT.println( steps );
      }

      IMUreading  = true;
    }

  } else {

    // Did the previous read finish?
    if (currentTime - lastIMUread > IMU_PERIOD+4000) { // nominal dt2

      IMUreading_t one;
      one.t        = lastIMUread;
      one.dt2      = currentTime - one.t;

      //  Yes, get the reading now.
      IMUreadings.write( one );

      IMUreading  = false;
    }
  }
} // readWIMU

//------------------------------------------------------

void readSteering()
{
  currentTime = micros();

  if (currentTime - lastSteeringRead >= STEERING_PERIOD) {

    SteeringReadings.write( currentTime );

    uint8_t steps = 0;
    do {
      lastSteeringRead += STEERING_PERIOD;
      steps++;
    } while (currentTime - lastSteeringRead >= STEERING_PERIOD);

    if (steps > 1) {// then sample missed?
      DEBUG_PORT.print( 'S' );
      DEBUG_PORT.print( currentTime );
      DEBUG_PORT.print( '=' );
      DEBUG_PORT.println( steps );
    }
  }
} // readSteering

//------------------------------------------------------

#ifdef SAVE_TO_SD

void yield()
{
  if (readDuringYield) {
    yieldCount++;

    readWSS();
    readSteering();
    readIMU();
    readGPS(); // simulated
  }
} // yield

#endif

//------------------------------------------------------

void saveData()
{
  bool     wroteSomething   = false;
  uint32_t startLoggingTime = micros();

  readDuringYield = true;

  while (WSSreadings.available()) {
    uint32_t t = WSSreadings.read();

    #ifdef SAVE_TO_SD
      if (saveFlag) {
        logfile.print( F("W,") );
        logfile.print( t );
        logfile.println( F(",0,0") );
        wroteSomething = true;
      }
    #endif
  }

  while (SteeringReadings.available()) {
    uint32_t t = SteeringReadings.read();

    #ifdef SAVE_TO_SD
      if (saveFlag) {
        logfile.print( F("S,") );
        logfile.print( t );
        logfile.println( F(",295") );
        wroteSomething = true;
      }
    #endif
  }


  while (IMUreadings.available()) {
    IMUreading_t one = IMUreadings.read();

    #ifdef SAVE_TO_SD
      if (saveFlag) {
        logfile.print( F("I,") );
        logfile.print( one.t );
        logfile.print( F(",70,20,104,87,15,16,5759,9,25,17,41,") );
        logfile.print( one.dt2 );
        logfile.print( ',' );
        logfile.println( one.t ); // ~dt3
        wroteSomething = true;
      }
    #endif
  }

  while (GPSreadings.available()) {
    uint32_t t = GPSreadings.read();

    #ifdef SAVE_TO_SD
      if (saveFlag) {
        logfile.print( F("G,") );
        logfile.print( t );
        logfile.println( F(",20:59:31.95,48.6082867,2.41341170,0,2050") );
        wroteSomething = true;
      }
    #endif
  }

  if (wroteSomething) {

    #ifdef SAVE_TO_SD
      uint32_t currentMillis = millis();
      if ( currentMillis - lastFlushTime > FLUSH_PERIOD ) {
        lastFlushTime = currentMillis;
        logfile.flush();
        flushCount++;
      }

      // Accumulate how long these SD write operations took
      currentTime  = micros();
      loggingTime += currentTime - startLoggingTime;

      // Occasionally write an 'L record
      if (currentTime - lastLrecord > LRECORD_PERIOD) {
        lastLrecord = currentTime; // It's ok if these times creep a little

        logfile.print( F("L,") );
        logfile.print( currentTime );
        logfile.print( ',' );
        logfile.print( loggingTime );
        logfile.print( ',' );
        logfile.print( flushCount );
        logfile.print( ',' );
        uint32_t oldYieldCount = yieldCount; // this can change in the next write
        logfile.println( oldYieldCount );

        yieldCount  -= oldYieldCount; // usually resets to 0
        flushCount   = 0;

        currentTime  = micros();
        loggingTime  = currentTime - lastLrecord; // this time gets added to *next* logging times
      }
    #endif
  }

  readDuringYield = false;
} // saveData

//------------------------------------------------------

#ifdef SAVE_TO_SD

void setupSDCard()
{
  char filename[15] = "data_##.txt";

  for (uint8_t i=0; i<100; i++) {
    filename[5] = '0' + i/10;
    filename[6] = '0' + i%10;
    if (!SD.exists(filename)) {
      // Use this one!
      break;
    }
  }

  logfile = SD.open(filename, FILE_WRITE);
  if (!logfile) {
    DEBUG_PORT.print( F("Couldn't create ") );
    DEBUG_PORT.println(filename);

    digitalWrite(LED,HIGH);  // leave the LED on so I know there is a problem

    while (true) {}
  }

  DEBUG_PORT.print( F("Writing to ") );
  DEBUG_PORT.println(filename);

  lastFlushTime = millis();

} // setupSDCard

#endif

//--------------------------------------------------------

void initReadings()
{
  currentTime = micros() - 350UL; // avoid "future" time for steering & IMU

  lastGPStime      = currentTime;
  while (GPSreadings.available())
    GPSreadings.read();

  #ifdef SAVE_TO_SD
    lastFlushTime  = currentTime;
    lastLrecord    = currentTime;
  #endif

  lastWSSread      = currentTime;            // Sample this first: pretty constant, very quick
  while (WSSreadings.available())
    WSSreadings.read();

  lastSteeringRead = currentTime + 100UL;      // Sample 2nd a little later, WSS takes ~100us
  while (SteeringReadings.available())
    SteeringReadings.read();

  lastIMUread      = currentTime + 100UL + 250UL; // Sample this last: finishes at random time
  while (IMUreadings.available())
    IMUreadings.read();

} // initReadings

//--------------------------------------------------------

void testButtonToSave()
{
  bool buttonState = saveFlag; // digitalRead(buttonPinToSave);

  // Instead of a button, accept a command from the DEBUG_PORT
  while (DEBUG_PORT.available()) {
    char c = DEBUG_PORT.read();
    if (c == '0')
      buttonState = false;
    else if (c == '1')
      buttonState = true;
  }

  if (saveFlag != buttonState) {
    saveFlag = buttonState;
    digitalWrite( LED, saveFlag );

    if (saveFlag) {
      #ifdef SAVE_TO_SD
        setupSDCard();
      #endif
      initReadings();

    } else { // !saveFlag
      #ifdef SAVE_TO_SD
        logfile.close();
      #endif
      DEBUG_PORT.println( F("logfile closed.") );
    }
  }

} // testButtonToSave

//-------------------------------------------------------

void setup()
{
  DEBUG_PORT.begin(115200);
  DEBUG_PORT.println( F("Press 1 to start logging, 0 to stop") );
  DEBUG_PORT.flush(); // wait for header line to go out

  #ifdef SAVE_TO_SD
    DEBUG_PORT.println( F("Initializing SD card...") );

    if (!SD.begin(chipSelect)) {
      DEBUG_PORT.println( F("  SD card failed, or not present") );
      while (true) {
        digitalWrite(LED,HIGH);
        delay(75);
        digitalWrite(LED,LOW);
        delay(75);
      }
    }

    DEBUG_PORT.println( F("  SD card initialized.") );
  #endif

  initReadings();

} // setup

//-------------------------------------------------------

void loop()
{
  testButtonToSave();

  readWSS();
  readSteering();
  readIMU();
  readGPS(); // simulated

  saveData();
} // loop
